/* button.c,v 1.13 2004/02/13 00:03:46 jenglish Exp
 * Copyright (c) 2003, Joe English
 *
 * Tile package: button, checkbutton, and radiobutton widgets.
 */

#include <string.h>
#include <tk.h>
#include "tkTheme.h"
#include "widget.h"

/*
 * Label resources
 */

typedef struct
{
    /*
     * Text element resources:
     */
    Tcl_Obj *foregroundObj;
    Tcl_Obj *paddingObj;
    Tcl_Obj *textObj;
    Tcl_Obj *textVariableObj;
    Tcl_Obj *fontObj;
    Tcl_Obj *underlineObj;
    Tcl_Obj *anchorObj;
    Tcl_Obj *widthObj;

    TraceHandle	*textVariableTrace;
    Tk_Image	image;

    /*
     * Image element resources:
     */
    Tcl_Obj *imageObj;

    /*
     * Compound label/image resources:
     */
    Tcl_Obj *compoundObj;

    /*
     * Compatibility/legacy resources:
     */
    int stateOption;

    /*
     * @@@ Not sure what to do with this one:
     */
    Tcl_Obj *reliefObj;
} LabelPart;

typedef struct
{
    WidgetCore	core;
    LabelPart	label;
} Label;

static Tk_OptionSpec LabelOptionSpecs[] =
{
    {TK_OPTION_STRING, "-text", "text", "Text", "",
	Tk_Offset(Label,label.textObj), -1, 
	0,0,GEOMETRY_CHANGED },
    {TK_OPTION_STRING, "-textvariable", "textVariable", "Variable", "",
	Tk_Offset(Label,label.textVariableObj), -1, 
	TK_OPTION_NULL_OK,0,GEOMETRY_CHANGED },
    {TK_OPTION_COLOR, "-foreground", "foreground", "Foreground", NULL,
	Tk_Offset(Label,label.foregroundObj), -1, 
	TK_OPTION_NULL_OK,0,0 },
    {TK_OPTION_FONT, "-font", "font", "Font", DEFAULT_FONT,
	Tk_Offset(Label,label.fontObj), -1, 
	0,0,GEOMETRY_CHANGED },
    {TK_OPTION_INT, "-underline", "underline", "Underline",
	"-1", Tk_Offset(Label,label.underlineObj), -1, 
	0,0,0 },
    {TK_OPTION_INT, "-width", "width", "Width",
	NULL, Tk_Offset(Label,label.widthObj), -1, 
	TK_OPTION_NULL_OK,0,GEOMETRY_CHANGED },

    /*
     * Image resources
     */
    {TK_OPTION_STRING, "-image", "image", "Image", NULL/*default*/,
	Tk_Offset(Label,label.imageObj), -1, 
	TK_OPTION_NULL_OK,0,GEOMETRY_CHANGED },

    /*
     * Compound label/image resources.
     */
    {TK_OPTION_STRING_TABLE, "-compound", "compound", "Compound",
	 "none", Tk_Offset(Label,label.compoundObj), -1,
	 0,(ClientData)TTKCompoundStrings,GEOMETRY_CHANGED },
    {TK_OPTION_ANCHOR, "-anchor", "anchor", "Anchor",
	DEFAULT_ANCHOR, Tk_Offset(Label,label.anchorObj), -1, 0, 0, 0},
    {TK_OPTION_STRING, "-padding", "padding", "Padding",
	NULL, Tk_Offset(Label,label.paddingObj), -1,
	TK_OPTION_NULL_OK,0,GEOMETRY_CHANGED},

    /*
     * Compatibility/legacy resources:
     */
    {TK_OPTION_STRING_TABLE, "-state", "state", "State",
	 "normal", -1, Tk_Offset(Label,label.stateOption),
	 0,(ClientData)TTKStateStrings,STATE_CHANGED },

    {TK_OPTION_RELIEF, "-relief", "relief", "Relief",
	NULL, Tk_Offset(Label,label.reliefObj), -1, TK_OPTION_NULL_OK,0,0 },

    {TK_OPTION_END, 0,0,0, NULL, -1,-1, 0, (ClientData)CoreOptionSpecs, 0}
};

/*------------------------------------------------------------------------
 * Label utilities
 */

/*
 * Variable trace procedure for label -textvariable resource:
 */
static int
LabelTextVariableChanged(
    Tcl_Interp *interp, 	/* interpreter containing variable. */
    void *clientData,		/* widget record pointer */
    const char *value)		/* variable's new value */
{
    Label *labelPtr = clientData;
    Tcl_Obj *newText = Tcl_NewStringObj(value, -1);

    Tcl_IncrRefCount(newText);
    Tcl_DecrRefCount(labelPtr->label.textObj);
    labelPtr->label.textObj = newText;

    WidgetChanged(&labelPtr->core, REDISPLAY_REQUIRED|RELAYOUT_REQUIRED);

    return TCL_OK;
}

static int
LabelInitialize(Tcl_Interp *interp, void *recordPtr)
{
    Label *labelPtr = recordPtr;
    labelPtr->label.textVariableTrace = 0;
    labelPtr->label.image = 0;
    return TCL_OK;
}

static void
LabelCleanup(void *recordPtr)
{
    Label *labelPtr = recordPtr;
    if (labelPtr->label.textVariableTrace)
	UntraceVariable(labelPtr->label.textVariableTrace);
    if (labelPtr->label.image)
	Tk_FreeImage(labelPtr->label.image);
}

static int
LabelConfigure(Tcl_Interp *interp, void *recordPtr, int mask)
{
    Label *labelPtr = recordPtr;
    Tcl_Obj *textVarName = labelPtr->label.textVariableObj;
    TraceHandle *vt;
    Tk_Image image = NULL;

    if (textVarName == NULL || *Tcl_GetString(textVarName) == '\0') {
	vt = 0;
    } else {
	vt = TraceVariable(interp, labelPtr->label.textVariableObj,
	    LabelTextVariableChanged, labelPtr);
	if (!vt) return TCL_ERROR;
    }

    if (labelPtr->label.imageObj) {
	const char *imageName = Tcl_GetString(labelPtr->label.imageObj);
	if (imageName) {
	    image = Tk_GetImage(interp, labelPtr->core.tkwin, imageName, 0,0);
	    if (!image)
		goto error;
	}
    }

    if (CoreConfigure(interp, recordPtr, mask) != TCL_OK) {
error:
	if (image) Tk_FreeImage(image);
	if (vt) UntraceVariable(vt);
	return TCL_ERROR;
    }

    if (labelPtr->label.textVariableTrace)
	UntraceVariable(labelPtr->label.textVariableTrace);
    labelPtr->label.textVariableTrace = vt;

    if (image != labelPtr->label.image) {
	if (labelPtr->label.image)
	    Tk_FreeImage(labelPtr->label.image);
	labelPtr->label.image = image;
    }

    if (mask & STATE_CHANGED) {
	CheckStateOption(&labelPtr->core, labelPtr->label.stateOption);
    }

    return TCL_OK;
}

static int
LabelPostConfigure(Tcl_Interp *interp, void *recordPtr, int mask)
{
    Label *labelPtr = recordPtr;
    int status = TCL_OK;

    if (labelPtr->label.textVariableTrace)
	status = FireTrace(labelPtr->label.textVariableTrace);

    return status;
}

static WidgetCommandSpec LabelCommands[] =
{
    { "configure",	WidgetConfigureCommand },
    { "cget",		WidgetCgetCommand },
    { "instate",	WidgetInstateCommand },
    { "state",  	WidgetStateCommand },
    { NULL, NULL }
};

WidgetSpec LabelWidgetSpec =
{
    "TLabel",			/* className */
    sizeof(Label),		/* recordSize */
    LabelOptionSpecs,		/* optionSpecs */
    LabelCommands,		/* subcommands */
    LabelInitialize,		/* initializeProc */
    LabelCleanup,		/* cleanupProc */
    LabelConfigure,		/* configureProc */
    LabelPostConfigure,		/* postConfigureProc */
    WidgetGetLayout, 		/* getLayoutProc */
    WidgetSize, 		/* sizeProc */
    WidgetDoLayout,		/* layoutProc */
    WidgetDisplay,		/* displayProc */
    WIDGET_SPEC_END		/* sentinel */
};


/*------------------------------------------------------------------------
 * Button widget implementation.
 *
 * This inherits just about everything from the label,
 * adds a new subcommand "invoke", and resources "-command" and "-default"
 */

typedef struct
{
    Tcl_Obj *commandObj;
    Tcl_Obj *defaultStateObj;
} ButtonPart;

typedef struct
{
    WidgetCore	core;
    LabelPart	label;
    ButtonPart	button;
} Button;

/*
 * Option specifications:
 */
static Tk_OptionSpec ButtonOptionSpecs[] =
{
    WIDGET_TAKES_FOCUS,

    {TK_OPTION_STRING, "-command", "command", "Command",
	"", Tk_Offset(Button, button.commandObj), -1, 0,0,0},
    {TK_OPTION_STRING_TABLE, "-default", "default", "Default",
	"disabled", Tk_Offset(Button, button.defaultStateObj), -1, 0,
	(ClientData) TTKDefaultStrings, 0},


    {TK_OPTION_END, 0,0,0, NULL, -1,-1, 0, (ClientData)LabelOptionSpecs, 0}
};

static int
ButtonInvokeCommand(
    Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[], void *recordPtr)
{
    Button *buttonPtr = recordPtr;
    if (objc > 2) {
	Tcl_WrongNumArgs(interp, 1, objv, "invoke");
	return TCL_ERROR;
    }
    if (buttonPtr->core.state & TTK_STATE_DISABLED)
	return TCL_OK;
    return Tcl_EvalObjEx(interp, buttonPtr->button.commandObj, TCL_EVAL_GLOBAL);
}

static WidgetCommandSpec ButtonCommands[] =
{
    { "configure",	WidgetConfigureCommand },
    { "cget",		WidgetCgetCommand },
    { "invoke",		ButtonInvokeCommand },
    { "instate",	WidgetInstateCommand },
    { "state",  	WidgetStateCommand },
    { NULL, NULL }
};

WidgetSpec ButtonWidgetSpec =
{
    "TButton",			/* className */
    sizeof(Button),		/* recordSize */
    ButtonOptionSpecs,		/* optionSpecs */
    ButtonCommands,		/* subcommands */
    LabelInitialize,		/* initializeProc */
    LabelCleanup,		/* cleanupProc */
    LabelConfigure,		/* configureProc */
    LabelPostConfigure,		/* postConfigureProc */
    WidgetGetLayout, 		/* getLayoutProc */
    WidgetSize, 		/* sizeProc */
    WidgetDoLayout,		/* layoutProc */
    WidgetDisplay,		/* displayProc */

    WIDGET_SPEC_END		/* sentinel */
};

/*------------------------------------------------------------------------
 * Checkbutton widget implementation.
 */
typedef struct
{
    Tcl_Obj *variableObj;
    Tcl_Obj *onValueObj;
    Tcl_Obj *offValueObj;
    Tcl_Obj *commandObj;

    TraceHandle *variableTrace;

} CheckbuttonPart;

typedef struct
{
    WidgetCore core;
    LabelPart label;
    CheckbuttonPart checkbutton;
} Checkbutton;

/*
 * Option specifications:
 */
static Tk_OptionSpec CheckbuttonOptionSpecs[] =
{
    WIDGET_TAKES_FOCUS,

    {TK_OPTION_STRING, "-variable", "variable", "Variable",
	"", Tk_Offset(Checkbutton, checkbutton.variableObj), -1, 0,0,0},
    {TK_OPTION_STRING, "-onvalue", "onValue", "OnValue",
	"1", Tk_Offset(Checkbutton, checkbutton.onValueObj), -1, 0,0,0},
    {TK_OPTION_STRING, "-offvalue", "offValue", "OffValue",
	"0", Tk_Offset(Checkbutton, checkbutton.offValueObj), -1, 0,0,0},
    {TK_OPTION_STRING, "-command", "command", "Command",
	"", Tk_Offset(Checkbutton, checkbutton.commandObj), -1, 0,0,0},

    {TK_OPTION_END, 0,0,0, NULL, -1,-1, 0, (ClientData)LabelOptionSpecs, 0}
};

/*
 * Variable trace procedure for checkbutton -variable resource:
 */
static int
CheckbuttonVariableChanged(
    Tcl_Interp *interp, 	/* Interpreter containing variable. */
    void *clientData,		/* widget record pointer */
    const char *value)		/* Variable's new value */
{
    Checkbutton *checkPtr = clientData;

    if (!strcmp(value, Tcl_GetString(checkPtr->checkbutton.onValueObj))) {
	WidgetChangeState(&checkPtr->core, TTK_STATE_SELECTED, 0);
    } else {
	WidgetChangeState(&checkPtr->core, 0, TTK_STATE_SELECTED);
    }

    return TCL_OK;
}

static int
CheckbuttonInitialize(Tcl_Interp *interp, void *recordPtr)
{
    Checkbutton *checkPtr = recordPtr;
    checkPtr->checkbutton.variableTrace = TraceVariable(
	interp, checkPtr->checkbutton.variableObj,
	CheckbuttonVariableChanged, checkPtr);

    if (!checkPtr->checkbutton.variableTrace) return TCL_ERROR;

    return LabelInitialize(interp, recordPtr);
}

static void
CheckbuttonCleanup(void *recordPtr)
{
    Checkbutton *checkPtr = recordPtr;
    UntraceVariable(checkPtr->checkbutton.variableTrace);
    checkPtr->checkbutton.variableTrace = 0;
    LabelCleanup(recordPtr);
}

static int
CheckbuttonConfigure(Tcl_Interp *interp, void *recordPtr, int mask)
{
    Checkbutton *checkPtr = recordPtr;
    TraceHandle *vt = TraceVariable(interp, checkPtr->checkbutton.variableObj,
	CheckbuttonVariableChanged, checkPtr);

    if (!vt) 
	return TCL_ERROR;

    if (LabelConfigure(interp, recordPtr, mask) != TCL_OK){
	UntraceVariable(vt);
	return TCL_ERROR;
    }

    UntraceVariable(checkPtr->checkbutton.variableTrace);
    checkPtr->checkbutton.variableTrace = vt;

    return TCL_OK;
}

static int
CheckbuttonPostConfigure(Tcl_Interp *interp, void *recordPtr, int mask)
{
    Checkbutton *checkPtr = recordPtr;
    int status = TCL_OK;

    if (checkPtr->checkbutton.variableTrace)
	status = FireTrace(checkPtr->checkbutton.variableTrace);
    if (status == TCL_OK && !WidgetDestroyed(&checkPtr->core))
	status = LabelPostConfigure(interp, recordPtr, mask);
    return status;
}

/*
 * Checkbutton 'invoke' subcommand:
 * 	Toggles the checkbutton state.
 */
static int
CheckbuttonInvokeCommand(
    Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[], void *recordPtr)
{
    Checkbutton *checkPtr = recordPtr;
    WidgetCore *corePtr = &checkPtr->core;
    Tcl_Obj *newValue;

    if (objc > 2) {
	Tcl_WrongNumArgs(interp, 1, objv, "invoke");
	return TCL_ERROR;
    }
    if (corePtr->state & TTK_STATE_DISABLED)
	return TCL_OK;

    /*
     * Toggle the selected state.
     */
    if (corePtr->state & TTK_STATE_SELECTED)
	newValue = checkPtr->checkbutton.offValueObj;
    else
	newValue = checkPtr->checkbutton.onValueObj;

    if (Tcl_ObjSetVar2(interp,
	    checkPtr->checkbutton.variableObj, NULL, newValue,
	    TCL_GLOBAL_ONLY|TCL_LEAVE_ERR_MSG)
	== NULL)
	return TCL_ERROR;

    if (WidgetDestroyed(corePtr))
	return TCL_ERROR;

    return Tcl_EvalObjEx(interp,
	checkPtr->checkbutton.commandObj, TCL_EVAL_GLOBAL);
}

static WidgetCommandSpec CheckbuttonCommands[] =
{
    { "configure",	WidgetConfigureCommand },
    { "cget",		WidgetCgetCommand },
    { "invoke",		CheckbuttonInvokeCommand },
    { "instate",	WidgetInstateCommand },
    { "state",  	WidgetStateCommand },
    /* MISSING: select, deselect, toggle */
    { NULL, NULL }
};

WidgetSpec CheckbuttonWidgetSpec =
{
    "TCheckbutton",		/* className */
    sizeof(Checkbutton),	/* recordSize */
    CheckbuttonOptionSpecs,	/* optionSpecs */
    CheckbuttonCommands,	/* subcommands */
    CheckbuttonInitialize,	/* initializeProc */
    CheckbuttonCleanup,		/* cleanupProc */
    CheckbuttonConfigure,	/* configureProc */
    CheckbuttonPostConfigure,	/* postConfigureProc */
    WidgetGetLayout, 		/* getLayoutProc */
    WidgetSize, 		/* sizeProc */
    WidgetDoLayout,		/* layoutProc */
    WidgetDisplay,		/* displayProc */

    WIDGET_SPEC_END		/* sentinel */
};

/*------------------------------------------------------------------------
 * Radiobutton widget implementation.
 */

typedef struct
{
    Tcl_Obj *variableObj;
    Tcl_Obj *valueObj;
    Tcl_Obj *commandObj;

    TraceHandle	*variableTrace;

} RadiobuttonPart;

typedef struct
{
    WidgetCore core;
    LabelPart label;
    RadiobuttonPart radiobutton;
} Radiobutton;

/*
 * Option specifications:
 */
static Tk_OptionSpec RadiobuttonOptionSpecs[] =
{
    WIDGET_TAKES_FOCUS,

    {TK_OPTION_STRING, "-variable", "variable", "Variable",
	"", Tk_Offset(Radiobutton, radiobutton.variableObj), -1, 0,0,0},
    {TK_OPTION_STRING, "-value", "Value", "Value",
	"1", Tk_Offset(Radiobutton, radiobutton.valueObj), -1, 0,0,0},
    {TK_OPTION_STRING, "-command", "command", "Command",
	"", Tk_Offset(Radiobutton, radiobutton.commandObj), -1, 0,0,0},

    {TK_OPTION_END, 0,0,0, NULL, -1,-1, 0, (ClientData)LabelOptionSpecs, 0}
};

/*
 * Variable trace procedure for radiobuttons.
 */
static int
RadiobuttonVariableChanged(
    Tcl_Interp *interp, 	/* Interpreter containing variable. */
    void *clientData,		/* widget record pointer */
    const char *value)		/* Variable's new value */
{
    Radiobutton *radioPtr = clientData;

    if (!strcmp(value, Tcl_GetString(radioPtr->radiobutton.valueObj))) {
	WidgetChangeState(&radioPtr->core, TTK_STATE_SELECTED, 0);
    } else {
	WidgetChangeState(&radioPtr->core, 0, TTK_STATE_SELECTED);
    }

    return TCL_OK;
}

static int
RadiobuttonInitialize(Tcl_Interp *interp, void *recordPtr)
{
    Radiobutton *radioPtr = recordPtr;

    radioPtr->radiobutton.variableTrace = TraceVariable(
	interp, radioPtr->radiobutton.variableObj,
	RadiobuttonVariableChanged, radioPtr);
    if (!radioPtr->radiobutton.variableTrace) 
	return TCL_ERROR;

    return LabelInitialize(interp, recordPtr);
}

static void
RadiobuttonCleanup(void *recordPtr)
{
    Radiobutton *radioPtr = recordPtr;
    UntraceVariable(radioPtr->radiobutton.variableTrace);
    radioPtr->radiobutton.variableTrace = 0;
    LabelCleanup(recordPtr);
}

static int
RadiobuttonConfigure(Tcl_Interp *interp, void *recordPtr, int mask)
{
    Radiobutton *radioPtr = recordPtr;
    TraceHandle *vt = TraceVariable(interp, radioPtr->radiobutton.variableObj,
	RadiobuttonVariableChanged, radioPtr);

    if (!vt) 
	return TCL_ERROR;

    if (LabelConfigure(interp, recordPtr, mask) != TCL_OK) {
	UntraceVariable(vt);
	return TCL_ERROR;
    }

    UntraceVariable(radioPtr->radiobutton.variableTrace);
    radioPtr->radiobutton.variableTrace = vt;

    return TCL_OK;
}

static int
RadiobuttonPostConfigure(Tcl_Interp *interp, void *recordPtr, int mask)
{
    Radiobutton *radioPtr = recordPtr;
    int status = TCL_OK;

    if (radioPtr->radiobutton.variableTrace)
	status = FireTrace(radioPtr->radiobutton.variableTrace);
    if (status == TCL_OK && !WidgetDestroyed(&radioPtr->core))
	status = LabelPostConfigure(interp, recordPtr, mask);
    return status;
}

/*
 * Radiobutton 'invoke' subcommand:
 * 	Sets the radiobutton -variable to the -value, evaluates the -command.
 */
static int
RadiobuttonInvokeCommand(
    Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[], void *recordPtr)
{
    Radiobutton *radioPtr = recordPtr;
    WidgetCore *corePtr = &radioPtr->core;

    if (objc > 2) {
	Tcl_WrongNumArgs(interp, 1, objv, "invoke");
	return TCL_ERROR;
    }
    if (corePtr->state & TTK_STATE_DISABLED)
	return TCL_OK;

    if (Tcl_ObjSetVar2(interp,
	    radioPtr->radiobutton.variableObj, NULL,
	    radioPtr->radiobutton.valueObj,
	    TCL_GLOBAL_ONLY|TCL_LEAVE_ERR_MSG)
	== NULL)
	return TCL_ERROR;

    if (WidgetDestroyed(corePtr))
	return TCL_ERROR;

    return Tcl_EvalObjEx(interp,
	radioPtr->radiobutton.commandObj, TCL_EVAL_GLOBAL);
}

static WidgetCommandSpec RadiobuttonCommands[] =
{
    { "configure",	WidgetConfigureCommand },
    { "cget",		WidgetCgetCommand },
    { "invoke",		RadiobuttonInvokeCommand },
    { "instate",	WidgetInstateCommand },
    { "state",  	WidgetStateCommand },
    /* MISSING: select, deselect */
    { NULL, NULL }
};

WidgetSpec RadiobuttonWidgetSpec =
{
    "TRadiobutton",		/* className */
    sizeof(Radiobutton),	/* recordSize */
    RadiobuttonOptionSpecs,	/* optionSpecs */
    RadiobuttonCommands,	/* subcommands */
    RadiobuttonInitialize,	/* initializeProc */
    RadiobuttonCleanup,		/* cleanupProc */
    RadiobuttonConfigure,	/* configureProc */
    RadiobuttonPostConfigure,	/* postConfigureProc */
    WidgetGetLayout, 		/* getLayoutProc */
    WidgetSize, 		/* sizeProc */
    WidgetDoLayout,		/* layoutProc */
    WidgetDisplay,		/* displayProc */
    WIDGET_SPEC_END		/* sentinel */
};

