/* frame.c,v 1.9 2004/02/15 22:37:10 jenglish Exp
 * Copyright (c) 2004, Joe English
 *
 * Tile widget set: frame and labelframe widgets
 *
 * Note to maintainers: UNDER NO CIRCUMSTANCES is anyone to add
 * the -colormap, -visual, or -container resources to this widget.
 * Thank you.
 */

#include <tk.h>

#include "tkTheme.h"
#include "widget.h"

typedef struct
{
    Tcl_Obj	*backgroundObj;
    Tcl_Obj	*reliefObj;
    Tcl_Obj	*paddingObj;
    Tcl_Obj	*borderWidthObj;
} FramePart;

typedef struct 
{
    WidgetCore	core;
    FramePart	frame;
} Frame;

static Tk_OptionSpec FrameOptionSpecs[] =
{
    {TK_OPTION_BORDER, "-background", "background", "Background", NULL,
	Tk_Offset(Frame,frame.backgroundObj), -1, 
	TK_OPTION_NULL_OK,0,0 },
    {TK_OPTION_RELIEF, "-relief", "relief", "Relief", NULL,
	Tk_Offset(Frame,frame.reliefObj), -1, 
	TK_OPTION_NULL_OK,0,0 },
    {TK_OPTION_PIXELS, "-borderwidth", "borderWidth", "BorderWidth", "0",
	Tk_Offset(Frame,frame.borderWidthObj), -1,
	0,0,GEOMETRY_CHANGED },
    {TK_OPTION_STRING, "-padding", "padding", "Padding", NULL,
	Tk_Offset(Frame,frame.paddingObj), -1, 
	TK_OPTION_NULL_OK,0,GEOMETRY_CHANGED },

    {TK_OPTION_END, 0,0,0, NULL, -1,-1, 0, (ClientData)CoreOptionSpecs, 0}
};

static WidgetCommandSpec FrameCommands[] =
{
    { "configure",	WidgetConfigureCommand },
    { "cget",		WidgetCgetCommand },
    { "instate",	WidgetInstateCommand },
    { "state",  	WidgetStateCommand },
    { NULL, NULL }
};

/*
 * FrameMargins --
 * 	Compute internal margins for a frame widget.
 * 	This includes the -borderWidth, plus any additional -padding.
 * 	The Labelframe widget adjusts margins further.
 *
 * 	POSSIBLE BUG: This may be inaccurate; maybe check the theme instead?
 */
static TTK_Padding FrameMargins(Frame *framePtr)
{
    TTK_Padding margins = TTK_UniformPadding(0);

    /* Check -padding:
     */
    if (framePtr->frame.paddingObj) {
	TTK_GetPaddingFromObj(NULL, 
		framePtr->core.tkwin, framePtr->frame.paddingObj, &margins);
    }

    /* Account for -borderWidth:
     */
    if (framePtr->frame.borderWidthObj) {
	int bw = 0;
	Tk_GetPixelsFromObj(NULL, 
		framePtr->core.tkwin, framePtr->frame.borderWidthObj, &bw);
	margins.left += bw;
	margins.top += bw;
	margins.right += bw;
	margins.bottom += bw;
    }

    return margins;
}

/* FrameSize procedure --
 * 	The frame doesn't request a size of its own,
 * 	but it does have an internal border.
 */
static int FrameSize(void *recordPtr, int *widthPtr, int *heightPtr)
{
    Frame *framePtr = recordPtr;
    TTK_Padding margins = FrameMargins(framePtr);
    TTK_SetMargins(framePtr->core.tkwin, margins);
    return 0;
}

/*
 * FrameConfigure -- configure hook.
 * 	Just does some additional error-checking.
 */

static int FrameConfigure(Tcl_Interp *interp, void *recordPtr, int mask)
{
    Frame *framePtr = recordPtr;

    /*
     * Make sure -padding resource, if present, is correct:
     */
    if (framePtr->frame.paddingObj) {
	TTK_Padding unused;
	if (TTK_GetPaddingFromObj(interp,
		    	framePtr->core.tkwin,
			framePtr->frame.paddingObj,
			&unused) != TCL_OK) {
	    return TCL_ERROR;
	}
    }

    return CoreConfigure(interp, recordPtr, mask);
}

/* public */
WidgetSpec FrameWidgetSpec =
{
    "TFrame",			/* className */
    sizeof(Frame),		/* recordSize */
    FrameOptionSpecs,		/* optionSpecs */
    FrameCommands,		/* subcommands */
    NullInitialize,		/* initializeProc */
    NullCleanup,		/* cleanupProc */
    FrameConfigure,		/* configureProc */
    NullPostConfigure,		/* postConfigureProc */
    WidgetGetLayout, 		/* getLayoutProc */
    FrameSize,			/* sizeProc */
    WidgetDoLayout,		/* layoutProc */
    WidgetDisplay,		/* displayProc */
    WIDGET_SPEC_END		/* sentinel */
};

/* ======================================================================
 * Labelframe widget:
 * ======================================================================
 */

typedef enum {
    LABELANCHOR_E, LABELANCHOR_EN, LABELANCHOR_ES,
    LABELANCHOR_N, LABELANCHOR_NE, LABELANCHOR_NW,
    LABELANCHOR_S, LABELANCHOR_SE, LABELANCHOR_SW,
    LABELANCHOR_W, LABELANCHOR_WN, LABELANCHOR_WS
} LabelAnchor;

static const char *labelAnchorStrings[] = {
    "e", "en", "es", "n", "ne", "nw", "s", "se", "sw", "w", "wn", "ws", NULL
};

/* TODO: -labelwidget resource */
typedef struct {
    LabelAnchor	labelAnchor;
    Tcl_Obj	*textObj;
    Tcl_Obj	*fontObj;
} LabelframePart;

typedef struct
{
    WidgetCore  	core;
    FramePart   	frame;
    LabelframePart	label;
} Labelframe;

static Tk_OptionSpec LabelframeOptionSpecs[] =
{
    {TK_OPTION_STRING_TABLE, "-labelanchor", "labelAnchor", "LabelAnchor",
	"nw", -1, Tk_Offset(Labelframe, label.labelAnchor),
        0, (ClientData) labelAnchorStrings, GEOMETRY_CHANGED},
    {TK_OPTION_STRING, "-text", "text", "Text", "",
	Tk_Offset(Labelframe,label.textObj), -1, 
	0,0,GEOMETRY_CHANGED }, 
    {TK_OPTION_FONT, "-font", "font", "Font", DEFAULT_FONT,
	Tk_Offset(Labelframe,label.fontObj), -1, 
	0,0,GEOMETRY_CHANGED },

    /*
     * Override Frame defaults:
     */
    {TK_OPTION_RELIEF, "-relief", "relief", "Relief", "groove",
	Tk_Offset(Labelframe,frame.reliefObj), -1, 
	TK_OPTION_NULL_OK,0,0 },
    {TK_OPTION_PIXELS, "-borderwidth", "borderWidth", "BorderWidth", "2",
	Tk_Offset(Labelframe,frame.borderWidthObj), -1,
	TK_OPTION_NULL_OK,0,GEOMETRY_CHANGED },
    {TK_OPTION_END, 0,0,0, NULL, -1,-1, 0, (ClientData)FrameOptionSpecs, 0}
};

/*
 * LabelframeSize --
 * 	Like the frame, this doesn't request a size of its own
 * 	but it does have internal padding and a minimum size.
 */
static int LabelframeSize(void *recordPtr, int *widthPtr, int *heightPtr)
{
    Labelframe *lframePtr = recordPtr;
    WidgetCore *corePtr = &lframePtr->core;
    TTK_Padding margins = FrameMargins((Frame*)recordPtr);
    TTK_LayoutNode *textNode = TTK_LayoutFindNode(corePtr->layout, "text");
    int labelWidth, labelHeight;

    if (!textNode) {
	/* Something is wrong -- bail
	 */
	return 0;
    }

    TTK_LayoutReqSize(textNode, recordPtr, corePtr->tkwin, corePtr->state,
	    &labelWidth, &labelHeight);

    /* Adjust margins based on label size and position:
     */
    switch (lframePtr->label.labelAnchor) {
	case LABELANCHOR_W: case LABELANCHOR_WN: case LABELANCHOR_WS:
	    margins.left += labelWidth;
	    break;
	case LABELANCHOR_E: case LABELANCHOR_EN: case LABELANCHOR_ES:
	    margins.right += labelWidth;
	    break;
	case LABELANCHOR_N: case LABELANCHOR_NE: case LABELANCHOR_NW:
	    margins.top += labelHeight;
	    break;
	case LABELANCHOR_S: case LABELANCHOR_SE: case LABELANCHOR_SW:
	    margins.bottom += labelHeight;
	    break;
    }

    TTK_SetMargins(corePtr->tkwin,margins);
    /* @@@ this isn't quite enough: */ 
    Tk_SetMinimumRequestSize(corePtr->tkwin, labelWidth, labelHeight);

    return 0;
}

static void LabelframeDoLayout(void *recordPtr)
{
    Labelframe *lframePtr = recordPtr;
    WidgetCore *corePtr = &lframePtr->core;
    TTK_LayoutNode 
	*textNode = TTK_LayoutFindNode(corePtr->layout, "text"),
	*borderNode = TTK_LayoutFindNode(corePtr->layout, "border");
    TTK_Box borderParcel, labelParcel;
    int lw, lh;			/* Label width and height */
    unsigned sticky;		/* stickiness flags for final label placement */
    int spacing;		/* Extra space between label and border */

    /*
     * Do base layout:
     */
    TTK_LayoutGeometry(
	    corePtr->layout, recordPtr, corePtr->tkwin, corePtr->state,
	    0,0, Tk_Width(corePtr->tkwin), Tk_Height(corePtr->tkwin));

    /*
     * Adjust border and label parcels:
     */
    if (!textNode || !borderNode) { /* Something is wrong -- bail */
	return;
    }

    borderParcel = TTK_LayoutNodeParcel(borderNode);
    labelParcel = borderParcel;

    Tk_GetPixelsFromObj(NULL, 
	    lframePtr->core.tkwin, lframePtr->frame.borderWidthObj, &spacing);
    spacing = spacing * 2 + 4;

    /*
     * Adjust border and compute label parcel:
     */

    TTK_LayoutReqSize(textNode, recordPtr, corePtr->tkwin, corePtr->state,
	    &lw, &lh);

    switch (lframePtr->label.labelAnchor) 
    {
	case LABELANCHOR_W: case LABELANCHOR_WN: case LABELANCHOR_WS:
	    labelParcel = TTK_PackBox(&labelParcel, lw, lh, TTK_SIDE_LEFT);
	    labelParcel.y += spacing;
	    labelParcel.height -= 2*spacing;
	    borderParcel.x += lw / 2;
	    borderParcel.width -= lw / 2;
	    break;
	case LABELANCHOR_E: case LABELANCHOR_EN: case LABELANCHOR_ES:
	    labelParcel = TTK_PackBox(&labelParcel, lw, lh, TTK_SIDE_RIGHT);
	    labelParcel.y += spacing;
	    labelParcel.height -= 2*spacing;
	    borderParcel.width -= lw/2;
	    break;
	case LABELANCHOR_N: case LABELANCHOR_NE: case LABELANCHOR_NW:
	    labelParcel = TTK_PackBox(&labelParcel, lw, lh, TTK_SIDE_TOP);
	    labelParcel.x += spacing;
	    labelParcel.width -= 2*spacing;
	    borderParcel.y += lh / 2;
	    borderParcel.height -= lh / 2;
	    break;
	case LABELANCHOR_S: case LABELANCHOR_SE: case LABELANCHOR_SW:
	    labelParcel = TTK_PackBox(&labelParcel, lw, lh, TTK_SIDE_BOTTOM);
	    labelParcel.x += spacing;
	    labelParcel.width -= 2*spacing;
	    borderParcel.height -= lh / 2;
	    break;
    }

    /*
     * Adjust label parcel according to labelanchor
     */
    sticky = 0;		/* center */
    switch (lframePtr->label.labelAnchor) 
    {
	default:		sticky = 0; break;
	case LABELANCHOR_EN:
	case LABELANCHOR_WN:	sticky = TTK_STICK_N; break;
	case LABELANCHOR_ES:
	case LABELANCHOR_WS:	sticky = TTK_STICK_S; break;
	case LABELANCHOR_NE:
	case LABELANCHOR_SE:	sticky = TTK_STICK_E; break;
	case LABELANCHOR_NW:
	case LABELANCHOR_SW:	sticky = TTK_STICK_W; break;
    }
    labelParcel = TTK_PlaceBox(labelParcel, lw, lh, sticky);

    TTK_LayoutNodeSetParcel(borderNode, borderParcel);
    TTK_LayoutNodeSetParcel(textNode, labelParcel);
}

/* public */
WidgetSpec LabelframeWidgetSpec =
{
    "TLabelframe",		/* className */
    sizeof(Labelframe),		/* recordSize */
    LabelframeOptionSpecs,	/* optionSpecs */
    FrameCommands,		/* subcommands */
    NullInitialize,		/* initializeProc */
    NullCleanup,		/* cleanupProc */
    FrameConfigure,		/* configureProc */
    NullPostConfigure,		/* postConfigureProc */
    WidgetGetLayout, 		/* getLayoutProc */
    LabelframeSize,		/* sizeProc */
    LabelframeDoLayout,		/* layoutProc */
    WidgetDisplay,		/* displayProc */
    WIDGET_SPEC_END		/* sentinel */
};

