/*
 * tkstate.c,v 1.3 2004/01/14 01:37:15 jenglish Exp
 *
 * Tk widget state utilities.
 *
 * Copyright (c) 2003 Joe English.  Freely redistributable.
 *
 */

#include <string.h>

#include <tk.h>
#include "tkTheme.h"

/*
 * Table of state names.  Must be kept in sync with TTK_STATE_*
 * #defines in tkTheme.h.
 */

static const char *widgetStateNames[] =
{
    "active",		/* Mouse cursor is over widget or element */
    "disabled",		/* Widget is disabled */
    "focus",		/* Widget has keyboard focus */
    "pressed",		/* Pressed or "armed" */
    "selected",		/* "on", "true", "current", etc. */
    NULL
};

/*------------------------------------------------------------------------
 * +++ WidgetStateSpec object type:
 *
 * The string representation consists of a list of state names,
 * each optionally prefixed by an exclamation point (!).
 *
 * The internal representation uses the upper half of the longValue
 * to store the mask and the lower half to store the match.
 * If we ever get more than 16 states, this will need to be reconsidered...
 */

static int  WidgetStateSpecSetFromAny(Tcl_Interp *interp, Tcl_Obj *obj);
/* static void WidgetStateSpecFreeIntRep(Tcl_Obj *); */
#define WidgetStateSpecFreeIntRep 0		/* not needed */
static void WidgetStateSpecDupIntRep(Tcl_Obj *, Tcl_Obj *);
static void WidgetStateSpecUpdateString(Tcl_Obj *);

/*static*/
struct Tcl_ObjType WidgetStateSpecObjType =
{
    "WidgetStateSpec",
    WidgetStateSpecFreeIntRep,
    WidgetStateSpecDupIntRep,
    WidgetStateSpecUpdateString,
    WidgetStateSpecSetFromAny
};

static void WidgetStateSpecDupIntRep(Tcl_Obj *srcPtr, Tcl_Obj *copyPtr)
{
    copyPtr->internalRep.longValue = srcPtr->internalRep.longValue;
    copyPtr->typePtr = &WidgetStateSpecObjType;
}

static int WidgetStateSpecSetFromAny(Tcl_Interp *interp, Tcl_Obj *objPtr)
{
    int status;
    int objc;
    Tcl_Obj **objv;
    int i;
    unsigned int mask;
    unsigned int match;

    mask = match = 0;

    status = Tcl_ListObjGetElements(interp, objPtr, &objc, &objv);
    if (status != TCL_OK)
	return status;

    for (i = 0; i < objc; ++i)
    {
	int length, on, j;
	char *stateName = Tcl_GetStringFromObj(objv[i], &length);

	if (*stateName == '!')
	{
	    ++stateName;
	    --length;
	    on = 0;
	}
	else
	{
	    on = 1;
	}

	for (j = 0; widgetStateNames[j] != 0; ++j)
	    if (strncmp(stateName, widgetStateNames[j], (size_t)length) == 0)
		break;

    	if (widgetStateNames[j] == 0) {
	    if (interp) {
		Tcl_ResetResult(interp);
		Tcl_AppendResult(interp, "Invalid state name ", stateName,NULL);
	    }
	    return TCL_ERROR;
	}

	mask |= (1<<j);
	if (on) match |= (1<<j);
    }

    objPtr->typePtr->freeIntRepProc(objPtr);
    objPtr->typePtr = &WidgetStateSpecObjType;
    objPtr->internalRep.longValue = (mask << 16) | match;

    return TCL_OK;
}

static void WidgetStateSpecUpdateString(Tcl_Obj *objPtr)
{
    unsigned int mask = (objPtr->internalRep.longValue & 0xFFFF0000) >> 16;
    unsigned int match = objPtr->internalRep.longValue & 0x0000FFFF;
    Tcl_DString result;
    int i, len;

    Tcl_DStringInit(&result);

    for (i=0; widgetStateNames[i] != NULL; ++i) {
	if (mask & (1<<i)) {
	    if (!(match & (1<<i)))
		Tcl_DStringAppend(&result, "!", 1);
	    Tcl_DStringAppend(&result, widgetStateNames[i], -1);
	    Tcl_DStringAppend(&result, " ", 1);
	}
    }

    len = Tcl_DStringLength(&result);
    if (len) {
	/* 'len' includes extra trailing ' ' */
	objPtr->bytes = Tcl_Alloc((unsigned)len);
	objPtr->length = len-1;
	strncpy(objPtr->bytes, Tcl_DStringValue(&result), (size_t)len-1);
	objPtr->bytes[len-1] = '\0';
    } else {
	/* empty string */
	objPtr->length = 0;
	objPtr->bytes = Tcl_Alloc(1);
	*objPtr->bytes = '\0';
    }

    Tcl_DStringFree(&result);
}

Tcl_Obj *TTK_NewWidgetStateSpecObj(unsigned int mask, unsigned int match)
{
    Tcl_Obj *objPtr = Tcl_NewObj();

    Tcl_InvalidateStringRep(objPtr);
    objPtr->typePtr = &WidgetStateSpecObjType;
    objPtr->internalRep.longValue = (mask << 16) | match;

    return objPtr;
}

int TTK_GetWidgetStateSpecFromObj(
    Tcl_Interp *interp,
    Tcl_Obj *objPtr,
    TTK_WidgetStateSpec *spec)
{
    if (objPtr->typePtr != &WidgetStateSpecObjType)
    {
	int status = WidgetStateSpecSetFromAny(interp, objPtr);
	if (status != TCL_OK)
	    return status;
    }

    spec->mask = (objPtr->internalRep.longValue & 0xFFFF0000) >> 16;
    spec->match = objPtr->internalRep.longValue & 0x0000FFFF;
    return TCL_OK;
}


/*
 * Tk_StateMapLookup --
 *
 * 	A state map is a paired list of StateSpec / value pairs.
 *	Returns the value corresponding to the first matching state
 *	specification, or NULL if not found or an error occurs.
 */
Tcl_Obj *TTK_StateMapLookup(
    Tcl_Interp *interp,		/* Where to leave error messages; may be NULL */
    Tcl_Obj *map,		/* State map */
    unsigned int state) 	/* State to look up */
{
    Tcl_Obj **specs;
    int nSpecs;
    int j, status;

    status = Tcl_ListObjGetElements(interp, map, &nSpecs, &specs);
    if (status != TCL_OK)
	return NULL;

    for (j = 0; j < nSpecs; j += 2)
    {
	TTK_WidgetStateSpec spec;
	status = TTK_GetWidgetStateSpecFromObj(interp, specs[j], &spec);
	if (status != TCL_OK)
	    return NULL;
	if (TTK_WidgetStateMatches(state, &spec))
	    return specs[j+1];
    }
    if (interp) {
	Tcl_ResetResult(interp);
	Tcl_AppendResult(interp, "No match in state map", NULL);
    }
    return NULL;
}

/* TTK_GetStateMapFromObj --
 * 	Returns a TTK_StateMap from a Tcl_Obj*.
 * 	Since a TTK_StateMap is just a specially-formatted Tcl_Obj, 
 * 	this basically just checks for errors.
 */
TTK_StateMap TTK_GetStateMapFromObj(
    Tcl_Interp *interp,		/* Where to leave error messages; may be NULL */
    Tcl_Obj *mapObj)		/* State map */
{
    Tcl_Obj **specs;
    int nSpecs;
    int j, status;

    status = Tcl_ListObjGetElements(interp, mapObj, &nSpecs, &specs);
    if (status != TCL_OK)
	return NULL;

    if (nSpecs % 2 != 0) {
	if (interp)
	    Tcl_SetResult(interp,
		    "State map must have an even number of elements",
		    TCL_STATIC);
	return 0;
    }

    for (j = 0; j < nSpecs; j += 2) {
	TTK_WidgetStateSpec spec;
	if (TTK_GetWidgetStateSpecFromObj(interp, specs[j], &spec) != TCL_OK)
	    return NULL;
    }

    return mapObj;
}

/*EOF*/
