/*
 * xpTheme.c,v 1.40 2004/02/02 04:19:02 jenglish Exp
 *
 * Rough draft of a Tk style engine which uses the Windows XP theme engine.
 * Adapted from Georgios Petasis' XP theme patch.
 *
 * Copyright (c) 2003 by Georgios Petasis, petasis@iit.demokritos.gr.
 * Copyright (c) 2003 by Joe English
 * Copyright (c) 2003 by Pat Thoyts
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * See also:
 *
 * <URL: http://msdn.microsoft.com/library/en-us/
 *  	shellcc/platform/commctls/userex/refentry.asp >
 */

#include <windows.h>
#include <tk.h>

#ifdef NO_PRIVATE_HEADERS
#include "gunk.h"
#else
#include <tkWinInt.h>
#endif

#include "tkTheme.h"

/*
 * This #if chunk makes it possible to build this library
 * using older versions of the Platform SDK that don't support
 * the uxtheme library.
 */
#if (WINVER < 0x0501)
#if _MSC_VER > 100
#pragma message("warning: XP theme is unavailable.")
#endif
int
XPTheme_Init(Tcl_Interp *interp)
{
    return TCL_OK; 
}
#else /* WINVER >= 0x0501 */

#include <uxtheme.h>
#include <tmschema.h>

typedef HTHEME  (STDAPICALLTYPE OpenThemeDataProc)(HWND hwnd,
		 LPCWSTR pszClassList);
typedef HRESULT (STDAPICALLTYPE CloseThemeDataProc)(HTHEME hTheme);
typedef HRESULT (STDAPICALLTYPE DrawThemeBackgroundProc)(HTHEME hTheme,
                 HDC hdc, int iPartId, int iStateId, const RECT *pRect,
                 OPTIONAL const RECT *pClipRect);
typedef HRESULT (STDAPICALLTYPE DrawThemeParentBackgroundProc)(HWND hwnd,
                 HDC hdc, OPTIONAL const RECT *prc);
typedef HRESULT (STDAPICALLTYPE DrawThemeEdgeProc)(HTHEME hTheme, HDC hdc,
		 int iPartId, int iStateId, const RECT *pDestRect,
		 UINT uEdge, UINT uFlags, RECT *pContentRect);
typedef HRESULT (STDAPICALLTYPE DrawThemeTextProc)(HTHEME hTheme, HDC hdc,
		 int iPartId, int iStateId, LPCWSTR pszText, int iCharCount,
		 DWORD dwTextFlags, DWORD dwTextFlags2, const RECT *pRect);
typedef HRESULT	(STDAPICALLTYPE GetThemeMarginsProc)(HTHEME, HDC,
		 int iPartId, int iStateId, int iPropId,
		 OPTIONAL RECT *prc, MARGINS *pMargins);
typedef HRESULT	(STDAPICALLTYPE GetThemePartSizeProc)(HTHEME,HDC,
		 int iPartId, int iStateId,
		 RECT *prc, enum THEMESIZE eSize, SIZE *psz);
typedef HRESULT (STDAPICALLTYPE GetThemeTextExtentProc)(HTHEME hTheme, HDC hdc,
		 int iPartId, int iStateId, LPCWSTR pszText, int iCharCount,
		 DWORD dwTextFlags, const RECT *pBoundingRect, RECT *pExtentRect);
typedef BOOL    (STDAPICALLTYPE IsThemeActiveProc)(VOID);

typedef struct
{
    OpenThemeDataProc			*OpenThemeData;
    CloseThemeDataProc			*CloseThemeData;
    DrawThemeBackgroundProc		*DrawThemeBackground;
    DrawThemeParentBackgroundProc	*DrawThemeParentBackground;
    DrawThemeEdgeProc			*DrawThemeEdge;
    DrawThemeTextProc		        *DrawThemeText;
    GetThemeMarginsProc			*GetThemeMargins;
    GetThemePartSizeProc		*GetThemePartSize;
    GetThemeTextExtentProc		*GetThemeTextExtent;
    IsThemeActiveProc                   *IsThemeActive;
} XPThemeProcs;

typedef struct
{
    HINSTANCE hlibrary;
    XPThemeProcs *procs;
} XPThemeData;

/*
 *----------------------------------------------------------------------
 *
 * LoadXPThemeProcs --
 *	Initialize XP theming support.
 *
 *	XP theme support is included in UXTHEME.DLL
 *	We dynamically load this DLL at runtime instead of linking
 *	to it at build-time.
 *
 * Returns:
 *	A pointer to an XPThemeProcs table if successful, NULL otherwise.
 */

static XPThemeProcs *
LoadXPThemeProcs(HINSTANCE *phlib)
{
    OSVERSIONINFO os;

    /*
     * We have to check whether we are running at least on Windows XP.
     * In order to determine this we call GetVersionEx directly, although
     * it would be a good idea to wrap it inside a function similar to
     * TkWinGetPlatformId...
     */
    ZeroMemory(&os, sizeof(os));
    os.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
    GetVersionEx(&os);
    if (os.dwMajorVersion >= 5 && os.dwMinorVersion >= 1) {
	/*
	 * We are running under Windows XP or a newer version.
	 * Load the library "uxtheme.dll", where the native widget
	 * drawing routines are implemented.
	 */
	HINSTANCE handle;
	*phlib = handle = LoadLibrary("uxtheme.dll");
	if (handle != 0)
	{
	    /*
	     * We have successfully loaded the library. Proceed in storing the
	     * addresses of the functions we want to use.
	     */
	    XPThemeProcs *procs = (XPThemeProcs*)ckalloc(sizeof(XPThemeProcs));
#define LOADPROC(name) \
	(0 != (procs->name = (name ## Proc *)GetProcAddress(handle, #name) ))

	    if (   LOADPROC(OpenThemeData)
		&& LOADPROC(CloseThemeData)
		&& LOADPROC(DrawThemeBackground)
		&& LOADPROC(DrawThemeParentBackground)
		&& LOADPROC(DrawThemeEdge)
		&& LOADPROC(DrawThemeText)
		&& LOADPROC(GetThemeMargins)
		&& LOADPROC(GetThemePartSize)
		&& LOADPROC(GetThemeTextExtent)
		&& LOADPROC(IsThemeActive)
	    )
	    {
		return procs;
	    }
#undef LOADPROC
	    ckfree((char*)procs);
	}
    }
    return 0;
}

/*
 * XPThemeDeleteProc --
 *
 *      Release any theme allocated resources.
 */

static void
XPThemeDeleteProc(void *clientData)
{
    XPThemeData *themeData = clientData;
    FreeLibrary(themeData->hlibrary);
    ckfree(clientData);
}

static int
XPThemeEnabled(TTK_Theme theme, void *clientData)
{
    XPThemeData *themeData = clientData;
    return themeData->procs->IsThemeActive();
}

/*
 * MakeRect --
 * 	Helper routine.  Returns a RECT data structure.
 */
static RECT
MakeRect(int x, int y, int width, int height)
{
    RECT rc;
    rc.top = y;
    rc.left = x;
    rc.bottom = y + height;
    rc.right = x + width;
    return rc;
}


/*
 * Table for looking up an integer index based on widget state:
 */
typedef struct
{
    int index;			/* Value to return if this entry matches */
    unsigned int onBits;	/* Bits which must be set */
    unsigned int offBits;	/* Bitw which must be cleared */
} StateMap;

static int
StateMapIndex(StateMap *map, unsigned int state)
{
    if (!map)
	return 0;

    for (;;) {
	if ((state & map->onBits) == map->onBits
		&& (~state & map->offBits) == map->offBits)
	    return map->index;
	++map;
    }
}

/*
 * Map Tk state bitmaps to XP style enumerated values.
 *
 * In Tk, the "defaulted" state is indicated by a resource,
 * not in the 'state' bitmask.  Use one of the reserved
 * bits instead:
 */
#define TTK_STATE_DEFAULT TTK_STATE_RESERVED1

/*
 * Pushbuttons (Tk: "Button")
 */
static StateMap pushbutton_statemap[] =
{
    { PBS_DISABLED, 	TTK_STATE_DISABLED, 0 },
    { PBS_PRESSED, 	TTK_STATE_PRESSED, 0 },
    { PBS_HOT,		TTK_STATE_ACTIVE, 0 },
    { PBS_DEFAULTED,	TTK_STATE_DEFAULT, 0 },
    { PBS_NORMAL, 	0, 0 }
};

/*
 * Checkboxes (Tk: "Checkbutton")
 *
 * Missing: CBS_MIXEDDISABLED CBS_MIXEDHOT CBS_MIXEDNORMAL CBS_MIXEDPRESSED
 */
static StateMap checkbox_statemap[] =
{
{CBS_CHECKEDDISABLED,	TTK_STATE_DISABLED|TTK_STATE_SELECTED, 0},
{CBS_CHECKEDPRESSED,	TTK_STATE_PRESSED|TTK_STATE_SELECTED, 0},
{CBS_CHECKEDHOT,	TTK_STATE_ACTIVE|TTK_STATE_SELECTED, 0},
{CBS_CHECKEDNORMAL,	TTK_STATE_SELECTED, 0},
{CBS_UNCHECKEDDISABLED,	TTK_STATE_DISABLED, TTK_STATE_SELECTED},
{CBS_UNCHECKEDPRESSED,	TTK_STATE_PRESSED, TTK_STATE_SELECTED},
{CBS_UNCHECKEDHOT,	TTK_STATE_ACTIVE, TTK_STATE_SELECTED},
{CBS_UNCHECKEDNORMAL,	0,0 }
};

/*
 * Radiobuttons:
 */
static StateMap radiobutton_statemap[] =
{
{RBS_CHECKEDDISABLED,	TTK_STATE_DISABLED|TTK_STATE_SELECTED, 0},
{RBS_CHECKEDPRESSED,	TTK_STATE_PRESSED|TTK_STATE_SELECTED, 0},
{RBS_CHECKEDHOT,	TTK_STATE_ACTIVE|TTK_STATE_SELECTED, 0},
{RBS_CHECKEDNORMAL,	TTK_STATE_SELECTED, 0},
{RBS_UNCHECKEDDISABLED,	TTK_STATE_DISABLED, TTK_STATE_SELECTED},
{RBS_UNCHECKEDPRESSED,	TTK_STATE_PRESSED, TTK_STATE_SELECTED},
{RBS_UNCHECKEDHOT,	TTK_STATE_ACTIVE, TTK_STATE_SELECTED},
{RBS_UNCHECKEDNORMAL,	0,0 }
};

/*
 * Groupboxes (tk: "frame")
 */
static StateMap groupbox_statemap[] =
{
{GBS_DISABLED,	TTK_STATE_DISABLED, 0},
{GBS_NORMAL,	0,0 }
};

/*
 * Scrollbars (Tk: "Scrollbar.[vh]thumb")
 */
static StateMap scrollbar_statemap[] =
{
    { SCRBS_DISABLED, 	TTK_STATE_DISABLED, 0 },
    { SCRBS_PRESSED, 	TTK_STATE_PRESSED, 0 },
    { SCRBS_HOT,	TTK_STATE_ACTIVE, 0 },
    { SCRBS_NORMAL, 	0, 0 }
};

static StateMap uparrow_statemap[] =
{
    { ABS_UPDISABLED,	TTK_STATE_DISABLED, 0 },
    { ABS_UPPRESSED, 	TTK_STATE_PRESSED, 0 },
    { ABS_UPHOT,	TTK_STATE_ACTIVE, 0 },
    { ABS_UPNORMAL, 	0, 0 }
};

static StateMap downarrow_statemap[] =
{
    { ABS_DOWNDISABLED,	TTK_STATE_DISABLED, 0 },
    { ABS_DOWNPRESSED, 	TTK_STATE_PRESSED, 0 },
    { ABS_DOWNHOT,	TTK_STATE_ACTIVE, 0 },
    { ABS_DOWNNORMAL, 	0, 0 }
};

static StateMap leftarrow_statemap[] =
{
    { ABS_LEFTDISABLED,	TTK_STATE_DISABLED, 0 },
    { ABS_LEFTPRESSED, 	TTK_STATE_PRESSED, 0 },
    { ABS_LEFTHOT,	TTK_STATE_ACTIVE, 0 },
    { ABS_LEFTNORMAL, 	0, 0 }
};

static StateMap rightarrow_statemap[] =
{
    { ABS_RIGHTDISABLED,TTK_STATE_DISABLED, 0 },
    { ABS_RIGHTPRESSED, TTK_STATE_PRESSED, 0 },
    { ABS_RIGHTHOT,	TTK_STATE_ACTIVE, 0 },
    { ABS_RIGHTNORMAL, 	0, 0 }
};

/*
 * Scale (Tk: "Scale.[vh]thumb")
 */
static StateMap scale_statemap[] =
{
    { TUS_DISABLED, 	TTK_STATE_DISABLED, 0 },
    { TUS_PRESSED, 	TTK_STATE_PRESSED, 0 },
    { TUS_HOT,		TTK_STATE_ACTIVE, 0 },
    { TUS_NORMAL, 	0, 0 }
};

static StateMap tabitem_statemap[] =
{
    { TIS_DISABLED,     TTK_STATE_DISABLED, 0 },
    { TIS_SELECTED,     TTK_STATE_SELECTED, 0 },
    { TIS_HOT,          TTK_STATE_ACTIVE,   0 },
    { TIS_FOCUSED,      TTK_STATE_FOCUS,    0 },
    { TIS_NORMAL,       0,                  0 },
};

/*
 *----------------------------------------------------------------------
 * +++ Element data:
 *
 * The following structure is passed as the 'clientData' pointer
 * to all elements in this theme.  It contains data relevant
 * to a single XP Theme "part" (@@@ should probably be called 'PartData').
 */
typedef struct
{
    /*
     * Static data, initialized when element is registered:
     */
    XPThemeProcs *procs;	/* Pointer to theme procedure table */
    LPCWSTR	className;	/* Windows window class name */
    int 	partId;		/* BP_PUSHBUTTON, BP_CHECKBUTTON, etc. */
    StateMap	*statemap;	/* Lookup table to map Tk states to XP states */

    /*
     * Dynamic data, allocated by InitElementData:
     */
    HTHEME	hTheme;
    HDC		hDC;
    HWND	hwnd;

    /* For TkWinDrawableReleaseDC: */
    Drawable	drawable;
    TkWinDCState dcState;


} ElementData;

static ElementData *
NewElementData(
    XPThemeProcs *procs, LPCWSTR className, int partId, StateMap *statemap)
{
    ElementData *elementData = (ElementData*)ckalloc(sizeof(ElementData));

    elementData->procs = procs;
    elementData->className = className;
    elementData->partId = partId;
    elementData->statemap = statemap;

    elementData->hTheme = elementData->hDC = 0;

    return elementData;
}

static void DestroyElementData(ElementData *elementData)
{
    ckfree((char*)elementData);
}

/*
 * InitElementData --
 * 	Looks up theme handle.  If Drawable argument is non-NULL,
 * 	also initializes DC.
 *
 * Returns:
 * 	1 on success, 0 on error.
 * 	Caller must later call FreeElementData() so this element
 * 	can be reused.
 */

static int
InitElementData(ElementData *elementData, Tk_Window tkwin, Drawable d)
{
    Window win = Tk_WindowId(tkwin);
    if (win == (Window)NULL)
	return 0;

    elementData->hwnd = Tk_GetHWND(win);
    elementData->hTheme = elementData->procs->OpenThemeData(
	elementData->hwnd, elementData->className);

    if (!elementData->hTheme)
	return 0;

    elementData->drawable = d;
    if (d != 0) {
	elementData->hDC = TkWinGetDrawableDC(Tk_Display(tkwin), d,
	    &elementData->dcState);
    }

    return 1;
}

static void
FreeElementData(ElementData *elementData)
{
    elementData->procs->CloseThemeData(elementData->hTheme);
    if (elementData->drawable != 0) {
	TkWinReleaseDrawableDC(
	    elementData->drawable, elementData->hDC, &elementData->dcState);
    }
}

/*----------------------------------------------------------------------
 * +++ Null element.  Does nothing; used as a stub.
 */

typedef struct {
    Tcl_Obj	*unused;
} NullElement;

static TTK_ElementOptionSpec NullElementOptions[] = { {NULL} };

static void
NullElementGeometry(
    void *clientData, void *elementRecord,
    Tk_Window tkwin, int *widthPtr, int *heightPtr, TTK_Padding *paddingPtr)
{
}

static void
NullElementDraw(
    void *clientData, void *elementRecord,
    Tk_Window tkwin, Drawable d,
    int x, int y, int width, int height,
    unsigned int state)
{
}

static TTK_ElementSpec NullElementSpec =
{
    TK_STYLE_VERSION_2,
    sizeof(NullElement),
    NullElementOptions,
    NullElementGeometry,
    NullElementDraw
};

/*----------------------------------------------------------------------
 * +++ Generic element implementation.
 * 
 * Used for elements which are handled entirely by the XP Theme API,
 * such as radiobutton and checkbutton indicators, scrollbar arrows, etc.
 */

static void
GenericElementGeometry(
    void *clientData, void *elementRecord,
    Tk_Window tkwin, int *widthPtr, int *heightPtr, TTK_Padding *paddingPtr)
{
    ElementData *elementData = clientData;
    HRESULT result;
    SIZE size;

    if (!InitElementData(elementData, tkwin, 0))
	return;

    result = elementData->procs->GetThemePartSize(
	elementData->hTheme,
	elementData->hDC,
	elementData->partId,
	StateMapIndex(elementData->statemap, 0),
	NULL /*RECT *prc*/,
	TS_TRUE,
	&size);

    if (SUCCEEDED(result)) {
	*widthPtr = size.cx;
	*heightPtr = size.cy;
    } 

#if BROKEN
/* This doesn't seem to work.  Has anyone ever used GetThemeMargins()?
*/
    MARGINS margins = {0,0,0,0};
    result = elementData->procs->GetThemeMargins(
	elementData->hTheme,
	elementData->hDC,
	elementData->partId,
	StateMapIndex(elementData->statemap,0),
	TMT_MARGINS, NULL, &margins);
    if (SUCCEEDED(result)) {
	paddingPtr->left = margins.cxLeftWidth;
	paddingPtr->right = margins. cxRightWidth;
	paddingPtr->top = margins.cyTopHeight;
	paddingPtr->bottom = margins.cyBottomHeight;
    }
#endif
}

static void
GenericElementDraw(
    void *clientData, void *elementRecord,
    Tk_Window tkwin, Drawable d,
    int x, int y, int width, int height,
    unsigned int state)
{
    ElementData *elementData = clientData;
    RECT rc = MakeRect(x,y,width, height);

    if (!InitElementData(elementData, tkwin, d))
	return;

    elementData->procs->DrawThemeBackground(
	elementData->hTheme,
	elementData->hDC,
	elementData->partId,
	StateMapIndex(elementData->statemap, state),
	&rc,
	NULL/*pContentRect*/);

    FreeElementData(elementData);
}

static TTK_ElementSpec GenericElementSpec =
{
    TK_STYLE_VERSION_2,
    sizeof(NullElement),
    NullElementOptions,
    GenericElementGeometry,
    GenericElementDraw
};

/*----------------------------------------------------------------------
 * +++ Oriented generic element implementation.
 * 
 * Used for elements which are handled entirely by the XP Theme API,
 * but have a different part per orientation (eg: PROGRESS or SCROLLBAR
 * parts).
 *
 * If you register a new OrientedElement you should add it into this map
 * here are register the horizontal part as we lookup the vertical part
 * from the map.
 */

typedef struct
{
    LPCWSTR className;
    int     horizontal;
    int     vertical;
} OrientationMapItem;

static OrientationMapItem OrientationMap[] = 
{
    { L"PROGRESS", PP_BAR, PP_BARVERT },
    { L"PROGRESS", PP_CHUNK, PP_CHUNKVERT },
    { L"SCROLLBAR", SBP_THUMBBTNHORZ, SBP_THUMBBTNVERT },
    { L"SCROLLBAR", SBP_UPPERTRACKHORZ, SBP_UPPERTRACKVERT },
    { L"TRACKBAR", TKP_TRACK, TKP_TRACKVERT },
    { L"TRACKBAR", TKP_THUMBTOP, TKP_THUMBRIGHT },
    { NULL, 0, 0}
};

typedef struct
{
    Tcl_Obj *orientObj;
} OrientedElement;

static TTK_ElementOptionSpec OrientedElementOptions[] =
{
    { "-orient", TK_OPTION_ANY, 
	Tk_Offset(OrientedElement, orientObj), "horizontal" },
    { NULL }
};

static int
GetVerticalPart(LPCWSTR className, int part)
{
    OrientationMapItem *item = OrientationMap;
    for (; item->className != NULL; item++) {
	if (wcscmp(className, item->className) == 0 
	    && part == item->horizontal) {
	    part = item->vertical;
	    break;
	}
    }
    return part;
}

static void
OrientedElementGeometry(
    void *clientData, void *elementRecord,
    Tk_Window tkwin, int *widthPtr, int *heightPtr, TTK_Padding *paddingPtr)
{
    OrientedElement *options = elementRecord;
    ElementData *elementData = clientData;
    int orient, part = elementData->partId;
    HRESULT result;
    SIZE size;

    if (!InitElementData(elementData, tkwin, 0))
	return;

    TTK_GetOrientationFromObj(NULL, options->orientObj, &orient);
    if (orient == TTK_ORIENT_VERTICAL) {
	part = GetVerticalPart(elementData->className, elementData->partId);
    }

    result = elementData->procs->GetThemePartSize(
	elementData->hTheme,
	elementData->hDC,
	part,
	StateMapIndex(elementData->statemap, 0),
	NULL /*RECT *prc*/,
	TS_TRUE,
	&size);

    if (SUCCEEDED(result)) {
	*widthPtr = size.cx;
	*heightPtr = size.cy;
    } 

    FreeElementData(elementData);
}

#if UNUSED
static void
OrientedElementDraw(
    void *clientData, void *elementRecord,
    Tk_Window tkwin, Drawable d,
    int x, int y, int width, int height,
    unsigned int state)
{
    OrientedElement *options = elementRecord;
    ElementData *elementData = clientData;
    int orient, part  = elementData->partId;
    RECT rc = MakeRect(x,y,width, height);

    if (!InitElementData(elementData, tkwin, d))
	return;

    TTK_GetOrientationFromObj(NULL, options->orientObj, &orient);
    if (orient == TTK_ORIENT_VERTICAL) {
	part = GetVerticalPart(elementData->className, elementData->partId);
    }

    elementData->procs->DrawThemeBackground(
	elementData->hTheme,
	elementData->hDC,
	part,
	StateMapIndex(elementData->statemap, state),
	&rc,
	NULL/*pContentRect*/);

    FreeElementData(elementData);
}
#endif

#if UNUSED
static TTK_ElementSpec OrientedElementSpec =
{
    TK_STYLE_VERSION_2,
    sizeof(OrientedElement),
    OrientedElementOptions,
    OrientedElementGeometry,
    OrientedElementDraw
};
#endif

/*----------------------------------------------------------------------
 * +++ Background element implementation.
 *	Note that in XP, the default indicator is part of the background.
 */

typedef struct {
    Tcl_Obj	*backgroundObj;
    Tcl_Obj	*reliefObj;
    Tcl_Obj	*defaultStateObj;
} BackgroundElement;

static TTK_ElementOptionSpec BackgroundElementOptions[] = {
    { "-background", TK_OPTION_BORDER,
	    Tk_Offset(BackgroundElement,backgroundObj), DEFAULT_BACKGROUND },
    { "-relief", TK_OPTION_RELIEF, Tk_Offset(BackgroundElement,reliefObj),
    	"flat" },
    { "-default", TK_OPTION_ANY, Tk_Offset(BackgroundElement,defaultStateObj),
    	"normal" },
    {NULL}
};

static void
BackgroundElementDraw(
    void *clientData, void *elementRecord,
    Tk_Window tkwin, Drawable d,
    int x, int y, int width, int height,
    unsigned int state)
{
    ElementData *elementData = clientData;
    BackgroundElement *bg = elementRecord;
    RECT rc = MakeRect(x,y,width, height);
    Tk_3DBorder backgroundPtr = NULL;
    int relief = TK_RELIEF_FLAT;
    int defaultState = TK_BUTTON_DEFAULT_DISABLED;

    backgroundPtr = Tk_Get3DBorderFromObj(tkwin, bg->backgroundObj);
    Tk_GetReliefFromObj(NULL, bg->reliefObj, &relief);
    TTK_GetButtonDefaultStateFromObj(NULL,bg->defaultStateObj,&defaultState);
    if (defaultState == TK_BUTTON_DEFAULT_ACTIVE)
	state |= TTK_STATE_DEFAULT;

    Tk_Fill3DRectangle(tkwin, d, backgroundPtr,
	0,0,Tk_Width(tkwin),Tk_Height(tkwin), 0, TK_RELIEF_FLAT);

    if (!InitElementData(elementData, tkwin, d))
	return;

    if (relief != TK_RELIEF_FLAT) {
	elementData->procs->DrawThemeBackground(
	    elementData->hTheme,
	    elementData->hDC,
	    elementData->partId,
	    StateMapIndex(elementData->statemap, state),
	    &rc,
	    NULL/*pContentRect*/);
    }

    FreeElementData(elementData);
}

static TTK_ElementSpec BackgroundElementSpec =
{
    TK_STYLE_VERSION_2,
    sizeof(BackgroundElement),
    BackgroundElementOptions,
    NullElementGeometry,
    BackgroundElementDraw
};

#if UNUSED
/*----------------------------------------------------------------------
 * +++ Border element implementation.
 * 	Note that the border is actually drawn as part of the background
 * 	for most widgets; see also ButtonBorderElement
 */

/*
 * ReliefToEdge --
 * 	Convert a Tk "relief" value into an Windows "edge" value.
 * 	NB: Caller must check for RELIEF_FLAT, which does not correspond
 * 	to any Win32 enum.
 */
static unsigned int
ReliefToEdge(int relief)
{
    switch (relief) {
	case TK_RELIEF_FLAT:   return BDR_RAISEDOUTER;	/* @@@ */
	case TK_RELIEF_RAISED: return EDGE_RAISED;
	case TK_RELIEF_SUNKEN: return EDGE_SUNKEN;
	case TK_RELIEF_GROOVE: return EDGE_BUMP;
	case TK_RELIEF_SOLID:  return BDR_RAISEDOUTER;
	case TK_RELIEF_RIDGE:  return EDGE_ETCHED;
	default:               return BDR_RAISEDOUTER;
    }
}

typedef struct {
    Tcl_Obj	*reliefObj;
    Tcl_Obj	*defaultStateObj;
} BorderElement;

static TTK_ElementOptionSpec BorderElementOptions[] = {
    { "-relief", TK_OPTION_RELIEF, Tk_Offset(BorderElement,reliefObj), "flat" },
    { "-default", TK_OPTION_ANY, Tk_Offset(BorderElement,defaultStateObj),
    	"normal" },
    {NULL}
};

static void
BorderElementGeometry(
    void *clientData, void *elementRecord,
    Tk_Window tkwin, int *widthPtr, int *heightPtr, TTK_Padding *paddingPtr)
{
}

static void
BorderElementDraw(
    void *clientData, void *elementRecord,
    Tk_Window tkwin, Drawable d,
    int x, int y, int width, int height,
    unsigned int state)
{
    ElementData *elementData = clientData;
    BorderElement *border = elementRecord;
    RECT rc = MakeRect(x,y,width, height);
    int relief = TK_RELIEF_FLAT;
    int defaultState;

    Tk_GetReliefFromObj(NULL, border->reliefObj, &relief);
    TTK_GetButtonDefaultStateFromObj(NULL,
	    border->defaultStateObj, &defaultState);

    if (defaultState == TK_BUTTON_DEFAULT_ACTIVE)
	state |= TTK_STATE_DEFAULT;

    if (!InitElementData(elementData, tkwin, d))
	return;

    if (relief != TK_RELIEF_FLAT) {
	elementData->procs->DrawThemeEdge(
	    elementData->hTheme,
	    elementData->hDC,
	    elementData->partId,
	    StateMapIndex(elementData->statemap, state),
	    &rc,
	    ReliefToEdge(relief),
	    BF_RECT,
	    NULL/*pContentRect*/);
    }

    FreeElementData(elementData);
}

static TTK_ElementSpec BorderElementSpec =
{
    TK_STYLE_VERSION_2,
    sizeof(BorderElement),
    BorderElementOptions,
    BorderElementGeometry,
    BorderElementDraw
};
#endif

/*----------------------------------------------------------------------
 * +++ ButtonBorder element:
 * 	Used for pushbuttons, radiobuttons, and checkbuttons:
 * 	geometry only, border drawing is part of the XP theme background.
 */

typedef struct {
    Tcl_Obj	*unused;
} ButtonBorderElement;

static TTK_ElementOptionSpec ButtonBorderElementOptions[] = { {NULL} };

static void
ButtonBorderElementGeometry(
    void *clientData, void *elementRecord,
    Tk_Window tkwin, int *widthPtr, int *heightPtr, TTK_Padding *paddingPtr)
{
    /* @@@ Possibly: use GetThemeMargins here? */
    paddingPtr->left = paddingPtr->top
	= paddingPtr->right = paddingPtr->bottom = 4;
}

static TTK_ElementSpec ButtonBorderElementSpec =
{
    TK_STYLE_VERSION_2,
    sizeof(ButtonBorderElement),
    ButtonBorderElementOptions,
    ButtonBorderElementGeometry,
    NullElementDraw
};

/*----------------------------------------------------------------------
 * +++ Highlight element:
 * Draw focus rectangle.
 */

static void HighlightElementGeometry(
    void *clientData, void *elementRecord,
    Tk_Window tkwin, int *widthPtr, int *heightPtr, TTK_Padding *paddingPtr)
{
    paddingPtr->top=paddingPtr->left=paddingPtr->right=paddingPtr->bottom = 1;
}

static void
HighlightElementDraw(
    void *clientData, void *elementRecord,
    Tk_Window tkwin, Drawable d,
    int x, int y, int width, int height,
    unsigned int state)
{
    ElementData *elementData = clientData;

    if (state & TTK_STATE_FOCUS) {
	RECT rc = MakeRect(x,y,width,height);
	if (!InitElementData(elementData, tkwin, d))
	    return;
    	DrawFocusRect(elementData->hDC, &rc);
	FreeElementData(elementData);
    }
}

static TTK_ElementSpec HighlightElementSpec =
{
    TK_STYLE_VERSION_2,
    sizeof(NullElement),
    NullElementOptions,
    HighlightElementGeometry,
    HighlightElementDraw
};

/*----------------------------------------------------------------------
 * +++ Scrollbar thumb element.
 */

static void
ThumbElementDraw(
    void *clientData, void *elementRecord,
    Tk_Window tkwin, Drawable d,
    int x, int y, int width, int height,
    unsigned int state)
{
    ElementData *elementData = clientData;
    OrientedElement *thumb = elementRecord;
    int orient = TTK_ORIENT_HORIZONTAL;
    int thumbPartId, gripperPartId;
    RECT rc;

    if (!InitElementData(elementData, tkwin, d))
	return;

    TTK_GetOrientationFromObj(NULL, thumb->orientObj, &orient);
    if (orient == TTK_ORIENT_VERTICAL) {
	thumbPartId = SBP_THUMBBTNVERT;
	gripperPartId = SBP_GRIPPERVERT;
    } else {
	thumbPartId = SBP_THUMBBTNHORZ;
	gripperPartId = SBP_GRIPPERHORZ;
    }

    rc = MakeRect(x, y, width, height);

    elementData->procs->DrawThemeBackground(elementData->hTheme,
	elementData->hDC, thumbPartId,
	StateMapIndex(elementData->statemap, state),
	&rc, NULL);

    elementData->procs->DrawThemeBackground(elementData->hTheme,
	elementData->hDC, gripperPartId,
	StateMapIndex(elementData->statemap, state),
	&rc, NULL);

    FreeElementData(elementData);
}

static TTK_ElementSpec ThumbElementSpec =
{
    TK_STYLE_VERSION_2,
    sizeof(OrientedElement),
    OrientedElementOptions,
    OrientedElementGeometry,
    ThumbElementDraw
};

/*
 *----------------------------------------------------------------------
 * Trough
 */


typedef struct
{
    Tcl_Obj *orientObj;
    Tcl_Obj *grooveObj;
    Tcl_Obj *borderWidthObj;
} TroughElement;

static TTK_ElementOptionSpec TroughElementOptions[] =
{
    { "-orient", TK_OPTION_STRING_TABLE,
	Tk_Offset(TroughElement,orientObj), "vertical" },
    { "-groovewidth", TK_OPTION_PIXELS,
	Tk_Offset(TroughElement,grooveObj), "-1" },
    { "-borderwidth", TK_OPTION_PIXELS,
	Tk_Offset(TroughElement,borderWidthObj), DEFAULT_BORDERWIDTH },
    { NULL }
};

static void
TroughElementGeometry(
    void *clientData, void *elementRecord,
    Tk_Window tkwin, int *widthPtr, int *heightPtr, TTK_Padding *paddingPtr)
{
    TroughElement *troughPtr = elementRecord;
    int borderWidth = 0;

    /* adjust padding to prevent child elements drawing over the edge */
    Tk_GetPixelsFromObj(NULL, tkwin, troughPtr->borderWidthObj, &borderWidth);
    paddingPtr->left = paddingPtr->right = paddingPtr->top = paddingPtr->bottom
	= borderWidth;
}

/*
 * This draws a thin trough in the center of the parcel.
 */
static void
TroughElementDraw(
    void *clientData, void *elementRecord,
    Tk_Window tkwin, Drawable d,
    int x, int y, int width, int height,
    unsigned int state)
{
    TroughElement *trough = elementRecord;
    ElementData *elementData = clientData;
    RECT rc = MakeRect(x,y,width, height);
    MARGINS margins;
    int orient, groove, part = elementData->partId;
    HRESULT hr = S_OK;

    if (!InitElementData(elementData, tkwin, d))
	return;

    TTK_GetOrientationFromObj(NULL, trough->orientObj, &orient);
    Tk_GetPixelsFromObj(NULL, tkwin, trough->grooveObj, &groove);
    if (orient == TTK_ORIENT_VERTICAL) {
	part = GetVerticalPart(elementData->className, part);
    }

    hr = elementData->procs->GetThemeMargins(
	elementData->hTheme,
	elementData->hDC,
	part,
	TRS_NORMAL, TMT_BORDERSIZE, NULL, &margins);

    if (groove != -1) {
	if (orient == TTK_ORIENT_HORIZONTAL) {
	    rc = MakeRect(x, y + (height/2) - (margins.cyTopHeight/2),
			    width - 2, margins.cyTopHeight);
	} else {
	    rc = MakeRect(x + (width/2) - (margins.cyTopHeight/2), y,
			    margins.cyTopHeight, height - 2);
	}
    }

    hr = elementData->procs->DrawThemeBackground(
	elementData->hTheme,
	elementData->hDC,
	part,
	TRS_NORMAL,
	&rc,
	NULL);

    FreeElementData(elementData);
}

static TTK_ElementSpec TroughElementSpec =
{
    TK_STYLE_VERSION_2,
    sizeof(TroughElement),
    TroughElementOptions,
    TroughElementGeometry,
    TroughElementDraw
};

/*
 *----------------------------------------------------------------------
 * Text
 */


typedef struct
{
    Tcl_Obj *textObj;
    Tcl_Obj *fontObj;
} TextElement;

static TTK_ElementOptionSpec TextElementOptions[] =
{
    { "-text", TK_OPTION_STRING,
	Tk_Offset(TextElement,textObj), "" },
    { "-font", TK_OPTION_FONT,
	Tk_Offset(TextElement,fontObj), DEFAULT_FONT },
    { NULL }
};

static void
TextElementGeometry(
    void *clientData, void *elementRecord,
    Tk_Window tkwin, int *widthPtr, int *heightPtr, TTK_Padding *paddingPtr)
{
    TextElement *element = elementRecord;
    ElementData *elementData = clientData;
    RECT rc = {0, 0};
    HRESULT hr = S_OK;

    if (!InitElementData(elementData, tkwin, 0))
	return;

    hr = elementData->procs->GetThemeTextExtent(
	    elementData->hTheme,
	    elementData->hDC,
	    elementData->partId,
	    StateMapIndex(elementData->statemap, 0),
	    Tcl_GetUnicode(element->textObj),
	    -1,
	    DT_LEFT,// | DT_BOTTOM | DT_NOPREFIX,
	    NULL,
	    &rc);

    if (SUCCEEDED(hr)) {
	*widthPtr = rc.right - rc.left;
	if (*widthPtr < 80) *widthPtr = 80;
	*heightPtr = rc.bottom - rc.top;
	if (*heightPtr < 20) *heightPtr = 20;
    }

    FreeElementData(elementData);
}

static void 
TextElementDraw(
    ClientData clientData, void *elementRecord,
    Tk_Window tkwin, Drawable d,
    int x, int y, int width, int height,
    unsigned int state)
{
    TextElement *element = elementRecord;
    ElementData *elementData = clientData;
    RECT rc = MakeRect(x, y, width, height);
    HRESULT hr = S_OK;

    if (!InitElementData(elementData, tkwin, d))
	return;

    hr = elementData->procs->DrawThemeText(
	    elementData->hTheme,
	    elementData->hDC,
	    elementData->partId,
	    StateMapIndex(elementData->statemap, state),
	    Tcl_GetUnicode(element->textObj),
	    -1,
	    DT_LEFT,// | DT_BOTTOM | DT_NOPREFIX,
	    (state & TTK_STATE_DISABLED) ? DTT_GRAYED : 0,
	    &rc);
    FreeElementData(elementData);
}

static TTK_ElementSpec TextElementSpec =
{
    TK_STYLE_VERSION_2,
    sizeof(TextElement),
    TextElementOptions,
    TextElementGeometry,
    TextElementDraw
};

/*----------------------------------------------------------------------
 * +++ Initialization routine:
 */

int XPTheme_Init(Tcl_Interp *interp)
{
    XPThemeData *themeData;
    XPThemeProcs *procs;
    HINSTANCE hlibrary;
    TTK_Theme themePtr, parentPtr;
    ElementData *clientData;

    procs = LoadXPThemeProcs(&hlibrary);
    if (!procs)
	return TCL_ERROR;

    /*
     * Create the new style engine.
     */
    parentPtr = TTK_GetTheme(interp, "winnative");
    themePtr = TTK_CreateTheme(interp, "xpnative", parentPtr);

    if (!themePtr)
        return TCL_ERROR;

    /*
     * Set theme data and cleanup proc
     */

    themeData = (XPThemeData *)ckalloc(sizeof(XPThemeData));
    themeData->procs = procs;
    themeData->hlibrary = hlibrary;

    TTK_SetThemeEnabledProc(themePtr, XPThemeEnabled, themeData);
    TTK_RegisterCleanup(interp, themeData, XPThemeDeleteProc);

    /*
     * Register any new elements required for this style engine.
     */

    clientData = NewElementData(procs, 
	L"BUTTON", BP_PUSHBUTTON, pushbutton_statemap);
    TTK_RegisterElementSpec(themePtr, 
    	"Button.background", &BackgroundElementSpec, clientData);
    TTK_RegisterElementSpec(themePtr,
    	"Button.border", &ButtonBorderElementSpec, clientData);

    TTK_RegisterElementSpec(themePtr, "highlight", 
	&HighlightElementSpec, clientData);

    TTK_RegisterElementSpec(themePtr,
    	"Checkbutton.background", &BackgroundElementSpec, clientData);
    TTK_RegisterElementSpec(themePtr,
    	"Checkbutton.border", &ButtonBorderElementSpec, clientData);

    TTK_RegisterElementSpec(themePtr,
    	"Radiobutton.background", &BackgroundElementSpec, clientData);
    TTK_RegisterElementSpec(themePtr,
    	"Radiobutton.border", &ButtonBorderElementSpec, clientData);

    /* Indicator elements */
    TTK_RegisterElementSpec(themePtr, "Checkbutton.indicator", &GenericElementSpec,
	NewElementData(procs, L"BUTTON", BP_CHECKBOX, checkbox_statemap));
    TTK_RegisterElementSpec(themePtr, "Radiobutton.indicator", &GenericElementSpec,
	NewElementData(procs, L"BUTTON", BP_RADIOBUTTON, radiobutton_statemap));

    /* Frame elements */
    clientData = NewElementData(procs,
    	L"BUTTON", BP_GROUPBOX, groupbox_statemap);
    TTK_RegisterElementSpec(themePtr,
    	"Frame.background", &NullElementSpec, clientData);
    TTK_RegisterElementSpec(themePtr,
    	"Frame.border", &BackgroundElementSpec, clientData);
    TTK_RegisterElementSpec(themePtr,
    	"Labelframe.background", &NullElementSpec, clientData);
    TTK_RegisterElementSpec(themePtr,
    	"Labelframe.border", &BackgroundElementSpec, clientData);
//    TTK_RegisterElementSpec(themePtr,
//    	"Labelframe.text", &TextElementSpec, clientData);

    /* Scrollbar elements */
    TTK_RegisterElementSpec(themePtr, "Scrollbar.border", &NullElementSpec, NULL);
    TTK_RegisterElementSpec(themePtr, "thumb", &ThumbElementSpec,
	NewElementData(procs,L"SCROLLBAR",SBP_THUMBBTNHORZ,scrollbar_statemap));
    TTK_RegisterElementSpec(
    	themePtr, "Vertical.Scrollbar.trough", &GenericElementSpec,
	NewElementData(procs, L"SCROLLBAR", SBP_UPPERTRACKVERT, scrollbar_statemap));
    TTK_RegisterElementSpec(
    	themePtr, "Horizontal.Scrollbar.trough", &GenericElementSpec,
	NewElementData(procs, L"SCROLLBAR", SBP_UPPERTRACKHORZ, scrollbar_statemap));

    /* Arrow elements */
    TTK_RegisterElementSpec(themePtr, "uparrow",  &GenericElementSpec,
	NewElementData(procs, L"SCROLLBAR", SBP_ARROWBTN, uparrow_statemap));
    TTK_RegisterElementSpec(themePtr, "downarrow",  &GenericElementSpec,
	NewElementData(procs, L"SCROLLBAR", SBP_ARROWBTN, downarrow_statemap));
    TTK_RegisterElementSpec(themePtr, "leftarrow",  &GenericElementSpec,
	NewElementData(procs, L"SCROLLBAR", SBP_ARROWBTN, leftarrow_statemap));
    TTK_RegisterElementSpec(themePtr, "rightarrow", &GenericElementSpec,
	NewElementData(procs, L"SCROLLBAR", SBP_ARROWBTN, rightarrow_statemap));

    /* Scale elements */
    TTK_RegisterElementSpec(
    	themePtr, "Horizontal.Scale.slider", &GenericElementSpec,
	NewElementData(procs,L"TRACKBAR", TKP_THUMBTOP, scale_statemap));
    TTK_RegisterElementSpec(
    	themePtr, "Vertical.Scale.slider", &GenericElementSpec,
	NewElementData(procs,L"TRACKBAR", TKP_THUMBRIGHT, scale_statemap));
    TTK_RegisterElementSpec(themePtr, "Scale.trough", &TroughElementSpec,
	NewElementData(procs, L"TRACKBAR", TKP_TRACK, scale_statemap));

    /* Progress elements */
    TTK_RegisterElementSpec(themePtr, "Progress.trough", &TroughElementSpec,
	NewElementData(procs, L"PROGRESS", PP_BAR, NULL));
    TTK_RegisterElementSpec(
    	themePtr, "Horizontal.Progress.bar", &GenericElementSpec,
	NewElementData(procs, L"PROGRESS", PP_CHUNK, NULL));
    TTK_RegisterElementSpec(
    	themePtr, "Vertical.Progress.bar", &GenericElementSpec,
	NewElementData(procs, L"PROGRESS", PP_CHUNKVERT, NULL));

    /* Notebook elements */
    TTK_RegisterElementSpec(themePtr, "Notebook.background", &GenericElementSpec,
	NewElementData(procs, L"TAB", TABP_BODY, NULL));
    TTK_RegisterElementSpec(themePtr, "client", &GenericElementSpec,
	NewElementData(procs, L"TAB", TABP_PANE, NULL));
    TTK_RegisterElementSpec(themePtr, "tab", &GenericElementSpec,
	NewElementData(procs, L"TAB", TABP_TABITEM, tabitem_statemap));

    Tcl_PkgProvide(interp, "tile::xpnative", "0.0.2");

    return TCL_OK;
}

#endif /* WINVER < 0x0501 */
/*EOF*/
